;
; Platonic Solids demo program - S. Yockey, March 2023
;
; 3d rotation code was adapted from a Simons' Basic program from 
; https://retro64.altervista.org/blog/3d-graphics-rotating-cube-with-freebasic-pc-and-simons-basic-c64/
; that I originally coded in Slang for the DTV64. The Prog8 package includes a spinning cube demo that
; does not require floating point, but since we are doing all of the rotation calculations prior to any
; plotting, this is a minor factor.
;
; This program is compatible with x16emu & Box16 - R41 & R42

%import syslib
%import floats
%import textio

main { 
    uword[61] vx 
    uword[61] vy
    
    ubyte np
    
    float tx
    float ty
    float tz
    
    uword tt
    
 
; flags for each type, so we only calculate point rotations 1 time
	bool c_f = false
	bool d_f = false
	bool o_f = false
	bool i_f = false
	bool b_f = false
		
; master flag for whether or not to calculate 3d transformation
	bool calculate = true	
       
    const float PHI = 1.618		;close enough
    const ubyte TETRAHEDRON = 1
    const ubyte OCTAHEDRON = 3
    const ubyte CUBE = 5
    const ubyte DODECAHEDRON = 7
    const ubyte ICOSAHEDRON = 9
    const ubyte BUCKYBALL = 11
    
; tweaked values to keep objects within the 'screen refresh' area
    const float L = 19
    const float L1 = 7
    
    float[] icosx = [0,0,0,0,L,-L,-L,L,PHI*L,PHI*L,-PHI*L,-PHI*L]
    float[] icosy = [-L,L,L,-L,-PHI*L,-PHI*L,PHI*L,PHI*L,0,0,0,0]
    float[] icosz = [PHI*L,PHI*L,-PHI*L,-PHI*L,0,0,0,0,L,-L,-L,L]
    
    float[] dodecx = [-L,-L,-L,-L,L,L,L,L,-L/PHI,L/PHI,-L/PHI,L/PHI,0,0,0,0,-L*PHI,-L*PHI,L*PHI,L*PHI]
    float[] dodecy = [-L,L,L,-L,-L,L,-L,L,0,0,0,0,-L*PHI,-L*PHI,L*PHI,L*PHI,-L/PHI,L/PHI,-L/PHI,L/PHI]
    float[] dodecz = [L,L,-L,-L,L,L,-L,-L,L*PHI,L*PHI,-L*PHI,-L*PHI,L/PHI,-L/PHI,L/PHI,-L/PHI,0,0,0,0]
    
    float[] octx = [-L,0,L,0,0,0,L,L]
    float[] octy = [0,L,0,-L,0,0,L,-L]
    float[] octz = [0,0,0,0,L,-L,L,L]
    
    float[] cubex = [-L,-L,L,L,-L,-L,L,L]
    float[] cubey = [-L,L,L,-L,-L,L,L,-L]
    float[] cubez = [-L,-L,-L,-L,L,L,L,L]
    
; buckyball breaks Prog8 limitation on array size!!!!! Workaround required    
    float[] buckx = [0,0,0,0,
					L1, L1, -L1, -L1,
					L1*3*PHI, L1*3*PHI, -L1*3*PHI, -L1*3*PHI,
					L1*2, L1*2, L1*2, L1*2,
					-L1*2, -L1*2, -L1*2, -L1*2,
					L1*(2*PHI+1), L1*(2*PHI+1), L1*(2*PHI+1), L1*(2*PHI+1),
					-L1*(2*PHI+1), -L1*(2*PHI+1), -L1*(2*PHI+1), -L1*(2*PHI+1),
					L1*PHI, L1*PHI, L1*PHI, L1*PHI, 
					-L1*PHI,-L1*PHI, -L1*PHI, -L1*PHI, 	
					L1, L1, L1, L1, 
					-L1,-L1, -L1, -L1,	
					L1*(2+PHI), L1*(2+PHI), L1*(2+PHI),	L1*(2+PHI),
					 -L1*(2+PHI),-L1*(2+PHI)]
															
	float[] buckx2 = [-L1*(2+PHI), -L1*(2+PHI), L1*2*PHI, L1*2*PHI, L1*2*PHI, L1*2*PHI, -L1*2*PHI, -L1*2*PHI, -L1*2*PHI,-L1*2*PHI]		
										
					
    float[] bucky = [L1,L1,-L1,-L1,
    				L1*3*PHI, -L1*3*PHI, L1*3*PHI, -L1*3*PHI,
    				0,0,0,0,
    				L1*(2*PHI+1), L1*(2*PHI+1), -L1*(2*PHI+1), -L1*(2*PHI+1),
    				L1*(2*PHI+1), L1*(2*PHI+1), -L1*(2*PHI+1), -L1*(2*PHI+1),
    				L1*PHI, L1*PHI, -L1*PHI, -L1*PHI,
    				L1*PHI, L1*PHI, -L1*PHI, -L1*PHI,
    				L1*2, L1*2, -L1*2, -L1*2, 
    				L1*2, L1*2, -L1*2, -L1*2,
    				L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI), -L1*(2+PHI),
    				L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI), -L1*(2+PHI),    
    				L1*2*PHI, L1*2*PHI, -L1*2*PHI, -L1*2*PHI,
    				L1*2*PHI, L1*2*PHI ]    				
    				    				
    float[] bucky2 = [-L1*2*PHI, -L1*2*PHI, L1, L1, -L1, -L1, L1, L1, -L1, -L1]
    				
    				
    float[] buckz = [L1*3*PHI, -L1*3*PHI, L1*3*PHI, -L1*3*PHI,
    				0,0,0,0,
    				L1, -L1, L1, -L1,
    				L1*PHI, -L1*PHI, L1*PHI, -L1*PHI,
    				L1*PHI, -L1*PHI, L1*PHI, -L1*PHI,
    				L1*2, -L1*2, L1*2, -L1*2,
    				L1*2, -L1*2, L1*2, -L1*2,
    				L1*(2*PHI+1), -L1*(2*PHI+1), L1*(2*PHI+1), -L1*(2*PHI+1),
    				L1*(2*PHI+1), -L1*(2*PHI+1), L1*(2*PHI+1), -L1*(2*PHI+1),    
    				L1*2*PHI, -L1*2*PHI, L1*2*PHI, -L1*2*PHI, 
    				L1*2*PHI, -L1*2*PHI, L1*2*PHI, -L1*2*PHI,
    				L1, -L1, L1, -L1, 
    				L1, -L1]
    				
    float[] buckz2 = [ L1,-L1, L1*(2+PHI), -L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI), L1*(2+PHI), -L1*(2+PHI)]	
    				

	sub start() {
		float c
		float s
		float xt
		float yt
		float zt
		float t
		float t1
		float s1
		float c1
		float[50] x
		float[10] x2
		float[50] y		
		float[10] y2
		float[50] z		
		float[10] z2		
				
		
		word fs
		float rx
		uword duration
		bool debug = false
						
		ubyte key = 0
		ubyte q
		ubyte body_type = OCTAHEDRON
		ubyte numpts
			
		fs = 200
		rx = -0.1
		
		repeat {
		
			void cx16.screen_mode(3, false) 	;40x25textmode
			txt.color2(5,0)
			txt.clear_screen()
			txt.plot(10,1)
			txt.print("the platonic solids")
			txt.plot(5,9)
    		txt.print("c) cube")
    		txt.plot(5,11)
    		txt.print("d) dodecahedron")
    		txt.plot(5,13)
    		txt.print("o) octahedron") 
    		txt.plot(5,15)
    		txt.print("b) buckyball")
			txt.plot(5, 17)
    		txt.print("i) icosahedron")
    		txt.plot(5,19)
    		txt.print("q) quit")    	
    	
    		while key != 'q' {
    			key = c64.GETIN()
    				
    			when key {
    				'q' -> { 
    					goto quit }
    					
    				'c' -> {
    					body_type = CUBE
    					break
    				}
    						
    				'd' -> {
    					body_type = DODECAHEDRON
    					break
    				}
    					 
    				'o' -> {
    					body_type = OCTAHEDRON
    					break
    				}
    					 
    				'i' -> {
    					body_type = ICOSAHEDRON
    					break
    				}
    				'b' -> {
    					body_type = BUCKYBALL
    					break
    				}    			
    			}
    		}
    						
    		when body_type {
    			DODECAHEDRON -> {
    				numpts = 20
    				cx16.memory_copy(&dodecx,&x,100)
    				cx16.memory_copy(&dodecy,&y,100)
					cx16.memory_copy(&dodecz,&z,100) } 
				
    			OCTAHEDRON -> {
    				numpts = 8
    				octy[1] = 30
    				octy[3] = -30
    				octy[6] = 30
    				octy[7] = -30
    				cx16.memory_copy(&octx,&x,40)
    				cx16.memory_copy(&octy,&y,40)
					cx16.memory_copy(&octz,&z,40) }				
						
    			CUBE -> {
    				numpts = 8
    				cx16.memory_copy(&cubex,&x,40)
    				cx16.memory_copy(&cubey,&y,40)
					cx16.memory_copy(&cubez,&z,40) }							
										
    			ICOSAHEDRON -> {
    				numpts = 12
    				cx16.memory_copy(&icosx,&x,60)
    				cx16.memory_copy(&icosy,&y,60)
					cx16.memory_copy(&icosz,&z,60) }										
														
    			BUCKYBALL -> {
    				numpts = 60			
    				cx16.memory_copy(&buckx,&x,250)
    				cx16.memory_copy(&bucky,&y,250)
					cx16.memory_copy(&buckz,&z,250) 
					cx16.memory_copy(&buckx2,&x+250,50)
    				cx16.memory_copy(&bucky2,&y+250,50)
					cx16.memory_copy(&buckz2,&z+250,50) } 														
						
			}
				
			txt.clear_screen()			
			txt.plot(1, 1)
			when body_type {
				BUCKYBALL -> {
					txt.print("buckyball - aka truncated icosahedron")
					txt.plot(5,3)
					txt.print("60 vertices")
					txt.plot(5,4)
					txt.print("90 edges")
					txt.plot(5,5)
					txt.print("32 faces")
					if not b_f {			
						txt.plot(1,10)
    					txt.print("err...the oracles refuse to hear us,") 
    					txt.plot(1,12)
    					txt.print("since the buckyball, aka truncated")				
    					txt.plot(1,14)
    					txt.print("icosahedron, is not a platonic solid.")
    					txt.plot(5,18)
    					txt.print("...this could take a while...") 
    					calculate = true }	
    				else {
    					calculate = false }
    			}
    			
    			CUBE -> {
					txt.print("cube - fancy name unknown")
					txt.plot(5,3)
					txt.print("8 vertices")
					txt.plot(5,4)
					txt.print("12 edges")
					txt.plot(5,5)
					txt.print("6 faces")    	
			
					if not c_f {			
    					calculate = true }	
    				else {
    					calculate = false }
					}  
			
    			DODECAHEDRON -> {
					txt.print("dodecahedron")
					txt.plot(5,3)
					txt.print("20 vertices")
					txt.plot(5,4)
					txt.print("30 edges")
					txt.plot(5,5)
					txt.print("12 faces") 
			 
					if not d_f {			
    					calculate = true }	
    				else {
    					calculate = false }			
			  	
				}  			
					
    			ICOSAHEDRON -> {
					txt.print("icosahedron")
					txt.plot(5,3)
					txt.print("12 vertices")
					txt.plot(5,4)
					txt.print("30 edges")
					txt.plot(5,5)
					txt.print("20 faces") 
			  
					if not i_f {			
    					calculate = true }	
    				else {
    					calculate = false }			
			 	
				}  					
			 	
    			OCTAHEDRON -> {
					txt.print("octahedron - aka diamond or rhombus")
					txt.plot(5,3)
					txt.print("6 vertices")
					txt.plot(5,4)
					txt.print("12 edges")
					txt.plot(5,5)
					txt.print("8 faces")  
			
					if not o_f {			
    					calculate = true }	
    				else {
    					calculate = false }			  	
					}   				
				
    	}
    
    	if body_type != BUCKYBALL and calculate {
    		txt.plot(1,10)
    		txt.print("please wait....") 
    		txt.plot(1,14)
    		txt.print("while we consult the platonic oracles.")				
 		}
    	
    	txt.plot(5,25)
    	txt.print("press 'space' to pause rotation") 
    	txt.plot(5,26)
    	txt.print("or 'm' to return to the menu") 
    	
    	if  not calculate {
    		sys.wait(180)
    		goto nocalc }
    	
; calculate 3d transformation
		for q in 0 to 64 {
	
			c = floats.cos(rx)
			s = floats.sin(rx)	
		
			for np in 1 to numpts {
				if np <= 50 {
					tx = x[np-1]
					ty = y[np-1]
					tz = z[np-1] }
				else {
					tx = x2[np-51]
					ty = y2[np-51]
					tz = z2[np-51]	}				
							
												
			yt = ty 
			ty = c*yt-(s*tz)
			tz = s*yt+(c*tz)	
			xt = tx
			tx = c*xt+(s*tz)	
			tz = -s*xt+(c*tz)
			xt = tx
			tx = xt*c-(ty*s)
			ty = xt*s+(ty*c)	
			t = tx*fs
			t1 = tz+fs		
			t1 = t/t1			
			t = t1+160		
			tt = t as uword		
			vx[np] = tt
				
			t = ty*fs
			t1 = tz+fs
			t1 = t/t1
			t = t1+100
			tt = t as uword
			vy[np] = tt
		
		 
		}

	;copy calculated points to banked ram
	
	cx16.rambank(body_type)
	cx16.memory_copy(&vx,$a000 + (q as uword * 121 as uword),121)	
	cx16.rambank(0)

	cx16.rambank(body_type+1)
	cx16.memory_copy(&vy,$a000 + (q as uword * 121 as uword),121)	
	cx16.rambank(0)	

	rx += 0.098     
	}
	
nocalc:
	
	;320x200x256
	gfx3.screen_mode(2)
	
	;clears screen and back buffer	
	init_buffers()
	
	q = 0
	while key != 'm' {
		c64.SETTIM(0, 0, 0)

		; restore saved points from banked ram to working arrays		
		cx16.rambank(body_type)
		cx16.memory_copy($a000 + (q as uword * 121 as uword),&vx,121)
		cx16.rambank(0)		
	
		cx16.rambank(body_type+1)
		cx16.memory_copy($a000 + (q as uword * 121 as uword),&vy,121)
		cx16.rambank(0)			

    	;clear drawing area only
    	gfx3.fillrect(121,261,76,76,0)
           
		when body_type {
			DODECAHEDRON -> draw_dodec()	
				
			OCTAHEDRON -> draw_oct()	
			
			CUBE -> draw_cube()
			
			ICOSAHEDRON -> draw_icosa()
				
			BUCKYBALL -> draw_bucky()
		}				
			    
		;copy onto visible screen
		sys.waitvsync()
		blit2(120,260,76,76)
			
		duration = c64.RDTIM16()
		key = c64.GETIN()
		
		if key == ' ' {
			getkey() }
			
		if q < 64 q += 1
		else q = 0
			
			
	} 
	
	;retore text mode
	gfx3.screen_mode(0)
	void cx16.screen_mode(3, false) 	;40x25textmode		
	
	txt.color2(1,0)	
	txt.clear_screen()
    txt.plot(5,12)
    floats.print_f(60 as float/duration as float)
    txt.print(" frames per second!\n")        
    txt.plot(5,14)
    txt.print("press any key to continue...")
    
    getkey()
    key = 0
    }
    
quit:
	
} 

sub draw_cube() {

    ; drawing to back buffer (offscreen)

	gfx3.line(vx[1],vy[1] as uword + 200,vx[2],vy[2] as uword + 200,1)
	gfx3.line(vx[2],vy[2] as uword + 200,vx[3],vy[3] as uword + 200,1)
	gfx3.line(vx[3],vy[3] as uword + 200,vx[4],vy[4] as uword + 200,1)
	gfx3.line(vx[4],vy[4] as uword + 200,vx[1],vy[1] as uword + 200,1)
	gfx3.line(vx[5],vy[5] as uword + 200,vx[6],vy[6] as uword + 200,1)
	gfx3.line(vx[6],vy[6] as uword + 200,vx[7],vy[7] as uword + 200,1)
	gfx3.line(vx[7],vy[7] as uword + 200,vx[8],vy[8] as uword + 200,1)
	gfx3.line(vx[8],vy[8] as uword + 200,vx[5],vy[5] as uword + 200,1)
	gfx3.line(vx[1],vy[1] as uword + 200,vx[5],vy[5] as uword + 200,1)
	gfx3.line(vx[4],vy[4] as uword + 200,vx[8],vy[8] as uword + 200,1)
	gfx3.line(vx[2],vy[2] as uword + 200,vx[6],vy[6] as uword + 200,1)
	gfx3.line(vx[3],vy[3] as uword + 200,vx[7],vy[7] as uword + 200,1) 
	c_f = true			
}

sub draw_oct() {
    ; drawing to back buffer (offscreen)
       
	gfx3.line(vx[1],vy[1] as uword + 200,vx[2],vy[2] as uword + 200,1)
	gfx3.line(vx[2],vy[2] as uword + 200,vx[3],vy[3] as uword + 200,1)
	gfx3.line(vx[3],vy[3] as uword + 200,vx[4],vy[4] as uword + 200,1)
	gfx3.line(vx[4],vy[4] as uword + 200,vx[1],vy[1] as uword + 200,1)
	gfx3.line(vx[1],vy[1] as uword + 200,vx[5],vy[5] as uword + 200,1)
	gfx3.line(vx[1],vy[1] as uword + 200,vx[6],vy[6] as uword + 200,1)
	gfx3.line(vx[2],vy[2] as uword + 200,vx[5],vy[5] as uword + 200,1)
	gfx3.line(vx[2],vy[2] as uword + 200,vx[6],vy[6] as uword + 200,1)
	gfx3.line(vx[3],vy[3] as uword + 200,vx[5],vy[5] as uword + 200,1)
	gfx3.line(vx[3],vy[3] as uword + 200,vx[6],vy[6] as uword + 200,1)
	gfx3.line(vx[4],vy[4] as uword + 200,vx[5],vy[5] as uword + 200,1)
	gfx3.line(vx[4],vy[4] as uword + 200,vx[6],vy[6] as uword + 200,1) 
	o_f = true		
}

sub draw_dodec() {
    ; drawing to back buffer (offscreen)

	gfx3.line(vx[8],vy[8] as uword + 200,vx[16],vy[16] as uword + 200,1)
	gfx3.line(vx[14],vy[14] as uword + 200,vx[7],vy[7] as uword + 200,1)
	gfx3.line(vx[3],vy[3] as uword + 200,vx[16],vy[16] as uword + 200,1)
	gfx3.line(vx[19],vy[19] as uword + 200,vx[5],vy[5] as uword + 200,1)
	gfx3.line(vx[5],vy[5] as uword + 200,vx[13],vy[13] as uword + 200,1)			
	gfx3.line(vx[13],vy[13] as uword + 200,vx[1],vy[1] as uword + 200,1)
	gfx3.line(vx[1],vy[1] as uword + 200,vx[9],vy[9] as uword + 200,1)
	gfx3.line(vx[9],vy[9] as uword + 200,vx[10],vy[10] as uword + 200,1)
	gfx3.line(vx[10],vy[10] as uword + 200,vx[5],vy[5] as uword + 200,1)		
	gfx3.line(vx[10],vy[10] as uword + 200,vx[6],vy[6] as uword + 200,1)	
	gfx3.line(vx[6],vy[6] as uword + 200,vx[20],vy[20] as uword + 200,1)		
	gfx3.line(vx[20],vy[20] as uword + 200,vx[19],vy[19] as uword + 200,1)  			     
	gfx3.line(vx[19],vy[19] as uword + 200,vx[7],vy[7] as uword + 200,1)	
	gfx3.line(vx[8],vy[8] as uword + 200,vx[12],vy[12] as uword + 200,1)
	gfx3.line(vx[12],vy[12] as uword + 200,vx[7],vy[7] as uword + 200,1)
	gfx3.line(vx[12],vy[12] as uword + 200,vx[11],vy[11] as uword + 200,1)
	gfx3.line(vx[11],vy[11] as uword + 200,vx[4],vy[4] as uword + 200,1)		
	gfx3.line(vx[4],vy[4] as uword + 200,vx[14],vy[14] as uword + 200,1)
	gfx3.line(vx[4],vy[4] as uword + 200,vx[17],vy[17] as uword + 200,1)
	gfx3.line(vx[17],vy[17] as uword + 200,vx[1],vy[1] as uword + 200,1)			
	gfx3.line(vx[17],vy[17] as uword + 200,vx[18],vy[18] as uword + 200,1)
	gfx3.line(vx[18],vy[18] as uword + 200,vx[2],vy[2] as uword + 200,1)	
	gfx3.line(vx[2],vy[2] as uword + 200,vx[15],vy[15] as uword + 200,1);
	gfx3.line(vx[15],vy[15] as uword + 200,vx[6],vy[6] as uword + 200,1) ;												
	gfx3.line(vx[15],vy[15] as uword + 200,vx[16],vy[16] as uword + 200,1);
	gfx3.line(vx[20],vy[20] as uword + 200,vx[8],vy[8] as uword + 200,1)	
	gfx3.line(vx[3],vy[3] as uword + 200,vx[18],vy[18] as uword + 200,1)				
	gfx3.line(vx[3],vy[3] as uword + 200,vx[11],vy[11] as uword + 200,1)			
	gfx3.line(vx[2],vy[2] as uword + 200,vx[9],vy[9] as uword + 200,1)				
	gfx3.line(vx[13],vy[13] as uword + 200,vx[14],vy[14] as uword + 200,1)
	d_f = true		
}

sub draw_icosa() {
    ; drawing to back buffer (offscreen)
   
	gfx3.line(vx[1],vy[1] as uword + 200,vx[12],vy[12] as uword + 200,1)
	gfx3.line(vx[12],vy[12] as uword + 200,vx[2],vy[2] as uword + 200,1)
	gfx3.line(vx[2],vy[2] as uword + 200,vx[1],vy[1] as uword + 200,1)
	gfx3.line(vx[1],vy[1] as uword + 200,vx[6],vy[6] as uword + 200,1)				
	gfx3.line(vx[6],vy[6] as uword + 200,vx[5],vy[5] as uword + 200,1)
	gfx3.line(vx[5],vy[5] as uword + 200,vx[1],vy[1] as uword + 200,1)				
	gfx3.line(vx[1],vy[1] as uword + 200,vx[9],vy[9] as uword + 200,1)
	gfx3.line(vx[9],vy[9] as uword + 200,vx[2],vy[2] as uword + 200,1)	
	gfx3.line(vx[2],vy[2] as uword + 200,vx[8],vy[8] as uword + 200,1)	
	gfx3.line(vx[8],vy[8] as uword + 200,vx[9],vy[9] as uword + 200,1)				
	gfx3.line(vx[9],vy[9] as uword + 200,vx[10],vy[10] as uword + 200,1)		
	gfx3.line(vx[10],vy[10] as uword + 200,vx[8],vy[8] as uword + 200,1)			
	gfx3.line(vx[8],vy[8] as uword + 200,vx[3],vy[3] as uword + 200,1)		
	gfx3.line(vx[3],vy[3] as uword + 200,vx[7],vy[7] as uword + 200,1)		
	gfx3.line(vx[7],vy[7] as uword + 200,vx[8],vy[8] as uword + 200,1)				
	gfx3.line(vx[7],vy[7] as uword + 200,vx[2],vy[2] as uword + 200,1)  			     
	gfx3.line(vx[7],vy[7] as uword + 200,vx[11],vy[11] as uword + 200,1)		
	gfx3.line(vx[11],vy[11] as uword + 200,vx[6],vy[6] as uword + 200,1)	
	gfx3.line(vx[6],vy[6] as uword + 200,vx[12],vy[12] as uword + 200,1)
	gfx3.line(vx[6],vy[6] as uword + 200,vx[4],vy[4] as uword + 200,1)	
	gfx3.line(vx[4],vy[4] as uword + 200,vx[5],vy[5] as uword + 200,1)				
	gfx3.line(vx[5],vy[5] as uword + 200,vx[10],vy[10] as uword + 200,1)
	gfx3.line(vx[10],vy[10] as uword + 200,vx[4],vy[4] as uword + 200,1)
	gfx3.line(vx[4],vy[4] as uword + 200,vx[3],vy[3] as uword + 200,1)						
	gfx3.line(vx[9],vy[9] as uword + 200,vx[5],vy[5] as uword + 200,1)
	gfx3.line(vx[12],vy[12] as uword + 200,vx[11],vy[11] as uword + 200,1)	
	gfx3.line(vx[12],vy[12] as uword + 200,vx[7],vy[7] as uword + 200,1)
	gfx3.line(vx[11],vy[11] as uword + 200,vx[4],vy[4] as uword + 200,1) 											
	gfx3.line(vx[11],vy[11] as uword + 200,vx[3],vy[3] as uword + 200,1)
	gfx3.line(vx[10],vy[10] as uword + 200,vx[3],vy[3] as uword + 200,1)	
	i_f = true
}

sub draw_bucky() {		; still a few duplicate lines here!
    ; drawing to back buffer (offscreen)   
    
 	gfx3.line(vx[2],vy[2] as uword + 200,vx[4],vy[4] as uword + 200,1)	
	gfx3.line(vx[4],vy[4] as uword + 200,vx[32],vy[32] as uword + 200,1)	
	gfx3.line(vx[32],vy[32] as uword + 200,vx[56],vy[56] as uword + 200,1)
	gfx3.line(vx[56],vy[56] as uword + 200,vx[54],vy[54] as uword + 200,1)							
	gfx3.line(vx[54],vy[54] as uword + 200,vx[30],vy[30] as uword + 200,1)
	gfx3.line(vx[30],vy[30] as uword + 200,vx[2],vy[2] as uword + 200,1)							
	gfx3.line(vx[2],vy[2] as uword + 200,vx[34],vy[34] as uword + 200,1)
	gfx3.line(vx[34],vy[34] as uword + 200,vx[58],vy[58] as uword + 200,1)	
	gfx3.line(vx[58],vy[58] as uword + 200,vx[60],vy[60] as uword + 200,1)	
	gfx3.line(vx[60],vy[60] as uword + 200,vx[36],vy[36] as uword + 200,1)							
	gfx3.line(vx[36],vy[36] as uword + 200,vx[4],vy[4] as uword + 200,1)		
	gfx3.line(vx[32],vy[32] as uword + 200,vx[40],vy[40] as uword + 200,1)					
	gfx3.line(vx[40],vy[40] as uword + 200,vx[16],vy[16] as uword + 200,1)			
	gfx3.line(vx[16],vy[16] as uword + 200,vx[48],vy[48] as uword + 200,1)				
	gfx3.line(vx[48],vy[48] as uword + 200,vx[24],vy[24] as uword + 200,1)							
	gfx3.line(vx[24],vy[24] as uword + 200,vx[56],vy[56] as uword + 200,1)  						     
	gfx3.line(vx[40],vy[40] as uword + 200,vx[44],vy[44] as uword + 200,1)		
	gfx3.line(vx[44],vy[44] as uword + 200,vx[36],vy[36] as uword + 200,1)	
	gfx3.line(vx[36],vy[36] as uword + 200,vx[4],vy[4] as uword + 200,1)
	gfx3.line(vx[60],vy[60] as uword + 200,vx[28],vy[28] as uword + 200,1)	
	gfx3.line(vx[28],vy[28] as uword + 200,vx[52],vy[52] as uword + 200,1)				
	gfx3.line(vx[52],vy[52] as uword + 200,vx[20],vy[20] as uword + 200,1)
	gfx3.line(vx[20],vy[20] as uword + 200,vx[44],vy[44] as uword + 200,1)
	gfx3.line(vx[20],vy[20] as uword + 200,vx[8],vy[8] as uword + 200,1)						
	gfx3.line(vx[8],vy[8] as uword + 200,vx[6],vy[6] as uword + 200,1)
	gfx3.line(vx[6],vy[6] as uword + 200,vx[16],vy[16] as uword + 200,1)	
	gfx3.line(vx[24],vy[24] as uword + 200,vx[10],vy[10] as uword + 200,1)	
	gfx3.line(vx[10],vy[10] as uword + 200,vx[22],vy[22] as uword + 200,1)	
	gfx3.line(vx[22],vy[22] as uword + 200,vx[54],vy[54] as uword + 200,1)	
	gfx3.line(vx[22],vy[22] as uword + 200,vx[46],vy[46] as uword + 200,1)	
	gfx3.line(vx[46],vy[46] as uword + 200,vx[14],vy[14] as uword + 200,1)	
	gfx3.line(vx[14],vy[14] as uword + 200,vx[38],vy[38] as uword + 200,1)	
	gfx3.line(vx[38],vy[38] as uword + 200,vx[30],vy[30] as uword + 200,1)	
	gfx3.line(vx[14],vy[14] as uword + 200,vx[5],vy[5] as uword + 200,1)	
	gfx3.line(vx[5],vy[5] as uword + 200,vx[7],vy[7] as uword + 200,1)	;*		
	gfx3.line(vx[7],vy[7] as uword + 200,vx[18],vy[18] as uword + 200,1)			
	gfx3.line(vx[18],vy[18] as uword + 200,vx[42],vy[42] as uword + 200,1)			
	gfx3.line(vx[42],vy[42] as uword + 200,vx[38],vy[38] as uword + 200,1)			
	gfx3.line(vx[42],vy[42] as uword + 200,vx[34],vy[34] as uword + 200,1)		
	gfx3.line(vx[18],vy[18] as uword + 200,vx[50],vy[50] as uword + 200,1)			
	gfx3.line(vx[50],vy[50] as uword + 200,vx[26],vy[26] as uword + 200,1)			
	gfx3.line(vx[26],vy[26] as uword + 200,vx[58],vy[58] as uword + 200,1)					
	gfx3.line(vx[26],vy[26] as uword + 200,vx[12],vy[12] as uword + 200,1)				
	gfx3.line(vx[12],vy[12] as uword + 200,vx[28],vy[28] as uword + 200,1)			;~half			
		
	gfx3.line(vx[12],vy[12] as uword + 200,vx[11],vy[11] as uword + 200,1)		
	gfx3.line(vx[10],vy[10] as uword + 200,vx[9],vy[9] as uword + 200,1)			
				
	gfx3.line(vx[11],vy[11] as uword + 200,vx[25],vy[25] as uword + 200,1)	
	gfx3.line(vx[25],vy[25] as uword + 200,vx[57],vy[57] as uword + 200,1)	
	gfx3.line(vx[57],vy[57] as uword + 200,vx[59],vy[59] as uword + 200,1)	
	gfx3.line(vx[59],vy[59] as uword + 200,vx[27],vy[27] as uword + 200,1)	
	gfx3.line(vx[27],vy[27] as uword + 200,vx[11],vy[11] as uword + 200,1)	
	gfx3.line(vx[57],vy[57] as uword + 200,vx[33],vy[33] as uword + 200,1)	
	gfx3.line(vx[33],vy[33] as uword + 200,vx[41],vy[41] as uword + 200,1)	
	gfx3.line(vx[41],vy[41] as uword + 200,vx[17],vy[17] as uword + 200,1)	
	gfx3.line(vx[17],vy[17] as uword + 200,vx[49],vy[49] as uword + 200,1)
	gfx3.line(vx[49],vy[49] as uword + 200,vx[25],vy[25] as uword + 200,1)	
	gfx3.line(vx[5],vy[5] as uword + 200,vx[13],vy[13] as uword + 200,1)	
	gfx3.line(vx[13],vy[13] as uword + 200,vx[45],vy[45] as uword + 200,1)
	gfx3.line(vx[45],vy[45] as uword + 200,vx[46],vy[46] as uword + 200,1)	
	gfx3.line(vx[45],vy[45] as uword + 200,vx[21],vy[21] as uword + 200,1)	
	gfx3.line(vx[21],vy[21] as uword + 200,vx[53],vy[53] as uword + 200,1)	
	gfx3.line(vx[53],vy[53] as uword + 200,vx[29],vy[29] as uword + 200,1)	
	gfx3.line(vx[29],vy[29] as uword + 200,vx[37],vy[37] as uword + 200,1)		
	gfx3.line(vx[41],vy[41] as uword + 200,vx[37],vy[37] as uword + 200,1)				
	gfx3.line(vx[1],vy[1] as uword + 200,vx[3],vy[3] as uword + 200,1)
	gfx3.line(vx[1],vy[1] as uword + 200,vx[33],vy[33] as uword + 200,1)
	gfx3.line(vx[33],vy[33] as uword + 200,vx[57],vy[57] as uword + 200,1)	
	gfx3.line(vx[3],vy[3] as uword + 200,vx[35],vy[35] as uword + 200,1)	
	gfx3.line(vx[3],vy[3] as uword + 200,vx[31],vy[31] as uword + 200,1)		
	gfx3.line(vx[31],vy[31] as uword + 200,vx[55],vy[55] as uword + 200,1)	
	gfx3.line(vx[53],vy[53] as uword + 200,vx[55],vy[55] as uword + 200,1)		
	gfx3.line(vx[29],vy[29] as uword + 200,vx[1],vy[1] as uword + 200,1)			
	gfx3.line(vx[59],vy[59] as uword + 200,vx[35],vy[35] as uword + 200,1)		
	gfx3.line(vx[43],vy[43] as uword + 200,vx[35],vy[35] as uword + 200,1)		
	gfx3.line(vx[31],vy[31] as uword + 200,vx[39],vy[39] as uword + 200,1)		
	gfx3.line(vx[39],vy[39] as uword + 200,vx[43],vy[43] as uword + 200,1)			
	gfx3.line(vx[43],vy[43] as uword + 200,vx[19],vy[19] as uword + 200,1)			
	gfx3.line(vx[39],vy[39] as uword + 200,vx[15],vy[15] as uword + 200,1)			
	gfx3.line(vx[19],vy[19] as uword + 200,vx[8],vy[8] as uword + 200,1)		
	gfx3.line(vx[15],vy[15] as uword + 200,vx[6],vy[6] as uword + 200,1)			
	gfx3.line(vx[15],vy[15] as uword + 200,vx[47],vy[47] as uword + 200,1)				
	gfx3.line(vx[51],vy[51] as uword + 200,vx[19],vy[19] as uword + 200,1)			
	gfx3.line(vx[51],vy[51] as uword + 200,vx[27],vy[27] as uword + 200,1)			
	gfx3.line(vx[47],vy[47] as uword + 200,vx[23],vy[23] as uword + 200,1)			
	gfx3.line(vx[55],vy[55] as uword + 200,vx[23],vy[23] as uword + 200,1)			
	gfx3.line(vx[37],vy[37] as uword + 200,vx[13],vy[13] as uword + 200,1)			
	gfx3.line(vx[7],vy[7] as uword + 200,vx[17],vy[17] as uword + 200,1)		
	gfx3.line(vx[9],vy[9] as uword + 200,vx[21],vy[21] as uword + 200,1)	
	gfx3.line(vx[9],vy[9] as uword + 200,vx[23],vy[23] as uword + 200,1)	
	gfx3.line(vx[19],vy[19] as uword + 200,vx[43],vy[43] as uword + 200,1)
	gfx3.line(vx[13],vy[13] as uword + 200,vx[45],vy[45] as uword + 200,1)
	gfx3.line(vx[47],vy[47] as uword + 200,vx[48],vy[48] as uword + 200,1)	
	gfx3.line(vx[49],vy[49] as uword + 200,vx[50],vy[50] as uword + 200,1)	
	gfx3.line(vx[51],vy[51] as uword + 200,vx[52],vy[52] as uword + 200,1)	
	b_f = true
}

	
sub init_buffers() {
        ; erase all vram
        cx16.VERA_CTRL = 0
        cx16.VERA_ADDR_L = 0
        cx16.VERA_ADDR_M = 0
        cx16.VERA_ADDR_H = %00010000        ; auto incr 1
        repeat $ffff
            cx16.VERA_DATA0 = 0
        repeat $f960
            cx16.VERA_DATA0 = 0
    }


;only copy what we need to, from back to front display buffer
sub blit2(uword x, uword y, uword width, ubyte height) {
	uword ix
	ubyte iy
	uword address = x + (y-200)*320

	for iy in 0 to height {
        cx16.VERA_CTRL = 0
        cx16.VERA_ADDR_L = lsb(address)
        cx16.VERA_ADDR_M = msb(address)
        cx16.VERA_ADDR_H = 0 | %00010000    
              
        cx16.VERA_CTRL = 1      
        cx16.VERA_ADDR_L = lsb(address)
        cx16.VERA_ADDR_M = msb(address)+ 250
        
        if y > 204 {
        	cx16.VERA_ADDR_H = 1 | %00010000 }     
        else {
        	cx16.VERA_ADDR_H = 0 | %00010000 }
        
        	for ix in 0 to width {
            	cx16.VERA_DATA0 = cx16.VERA_DATA1 }
            	
	address += 320
	}
}
	
	
sub getkey() {
	ubyte key = 0
	while key == 0 {
		key = c64.GETIN() }
	}
	
} ;end of main section

; copy of Prog8 gfx2 library
; modified by me to support 320x200 256c mode and double buffering (as mode 2)
; Bitmap pixel graphics routines for the CommanderX16
; Custom routines to use the full-screen 640x480 and 320x240 screen modes.
; (These modes are not supported by the documented GRAPH_xxxx kernal routines)
;
; No text layer is currently shown, text can be drawn as part of the bitmap itself.
; Note: for similar graphics routines that also work on the C-64, use the "graphics" module instead.
; Note: for color palette manipulation, use the "palette" module or write Vera registers yourself.
; Note: this library implements code for various resolutions and color depths. This takes up memory.
;       If you're memory constrained you should probably not use this built-in library,
;       but make a copy in your project only containing the code for the required resolution.
;
;
; SCREEN MODE LIST:
;   mode 0 = reset back to default text mode
;   mode 1 = bitmap 320 x 240 monochrome
;   mode 2 = bitmap 320 x 200 x 4c
;   mode 3 = bitmap 320 x 240 x 16c (TODO not yet implemented)
;   mode 4 = bitmap 320 x 240 x 256c  (like SCREEN $80 but using this api instead of kernal)
;   mode 5 = bitmap 640 x 480 monochrome
;   mode 6 = bitmap 640 x 480 x 4c
;   higher color dephts in highres are not supported due to lack of VRAM


gfx3 {

    ; read-only control variables:
    ubyte active_mode = 0
    uword width = 0
    uword height = 0
    ubyte bpp = 0
    bool monochrome_dont_stipple_flag = false            ; set to false to enable stippling mode in monochrome displaymodes

    sub screen_mode(ubyte mode) {
        when mode {
            1 -> {
                ; lores monochrome
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 64
                cx16.VERA_DC_VSCALE = 64
                cx16.VERA_L1_CONFIG = %00000100
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = 0
                width = 320
                height = 240
                bpp = 1
            }
            2 -> {
                ; 320x200 256c
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 64
                cx16.VERA_DC_VSCALE = 64
                cx16.VERA_L1_CONFIG = %00000111
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = 0
        		const ubyte vstart = 20
        		const ubyte vheight = 200
        		cx16.VERA_CTRL = %00000010
        		cx16.VERA_DC_VSTART = vstart
        		cx16.VERA_DC_VSTOP = vstart + vheight - 1                
                
                width = 320
                height = 200
                bpp = 8
            }            
            
            ; TODO modes 2, 3 not yet implemented
            4 -> {
                ; lores 256c
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 64
                cx16.VERA_DC_VSCALE = 64
                cx16.VERA_L1_CONFIG = %00000111
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = 0
                width = 320
                height = 240
                bpp = 8
            }
            5 -> {
                ; highres monochrome
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 128
                cx16.VERA_DC_VSCALE = 128
                cx16.VERA_L1_CONFIG = %00000100
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = %00000001
                width = 640
                height = 480
                bpp = 1
            }
            6 -> {
                ; highres 4c
                cx16.VERA_DC_VIDEO = (cx16.VERA_DC_VIDEO & %11001111) | %00100000      ; enable only layer 1
                cx16.VERA_DC_HSCALE = 128
                cx16.VERA_DC_VSCALE = 128
                cx16.VERA_L1_CONFIG = %00000101
                cx16.VERA_L1_MAPBASE = 0
                cx16.VERA_L1_TILEBASE = %00000001
                width = 640
                height = 480
                bpp = 2
            }
            else -> {
                ; back to default text mode and colors
                cx16.VERA_CTRL = %10000000      ; reset VERA and palette
                c64.CINT()      ; back to text mode
                width = 0
                height = 0
                bpp = 0
                mode = 0
            }
        }

        active_mode = mode
        if bpp
            clear_screen()
    }
    
    sub clear_screen() {
        monochrome_stipple(false)
        position(0, 0)
        when active_mode {
            1 -> {
                ; lores monochrome
                repeat 240/2/8
                    cs_innerloop640()
            }
            ; TODO mode 2, 3
            2 -> {
                ; 320x200 256c
                repeat 200/2
                    cs_innerloop640()            
            }
            4 -> {
                ; lores 256c
                repeat 240/2
                    cs_innerloop640()
            }
            5 -> {
                ; highres monochrome
                repeat 480/8
                    cs_innerloop640()
            }
            6 -> {
                ; highres 4c
                repeat 480/4
                    cs_innerloop640()
            }
            ; modes 7 and 8 not supported due to lack of VRAM
        }
        position(0, 0)
    }


    sub monochrome_stipple(bool enable) {
        monochrome_dont_stipple_flag = not enable
    }

    sub rect(uword x, uword y, uword rwidth, uword rheight, ubyte color) {
        if rwidth==0 or rheight==0
            return
        horizontal_line(x, y, rwidth, color)
        if rheight==1
            return
        horizontal_line(x, y+rheight-1, rwidth, color)
        vertical_line(x, y+1, rheight-2, color)
        if rwidth==1
            return
        vertical_line(x+rwidth-1, y+1, rheight-2, color)
    }

    sub fillrect(uword x, uword y, uword rwidth, uword rheight, ubyte color) {
        if rwidth==0
            return
        repeat rheight {
            horizontal_line(x, y, rwidth, color)
            y++
        }
    }

    sub horizontal_line(uword x, uword y, uword length, ubyte color) {
        if length==0
            return
        when active_mode {
            1, 5 -> {
                ; monochrome modes, either resolution
                ubyte separate_pixels = (8-lsb(x)) & 7
                if separate_pixels as uword > length
                    separate_pixels = lsb(length)
                repeat separate_pixels {
                    ; TODO optimize this by writing a masked byte in 1 go
                    plot(x, y, color)
                    x++
                }
                length -= separate_pixels
                if length {
                    position(x, y)
                    separate_pixels = lsb(length) & 7
                    x += length & $fff8
                    %asm {{
                        lsr  length+1
                        ror  length
                        lsr  length+1
                        ror  length
                        lsr  length+1
                        ror  length
                        lda  color
                        bne  +
                        ldy  #0     ; black
                        bra  _loop
+                       lda  monochrome_dont_stipple_flag
                        beq  _stipple
                        ldy  #255       ; don't stipple
                        bra  _loop
_stipple                lda  y
                        and  #1         ; determine stipple pattern to use
                        bne  +
                        ldy  #%01010101
                        bra  _loop
+                       ldy  #%10101010
_loop                   lda  length
                        ora  length+1
                        beq  _done
                        sty  cx16.VERA_DATA0
                        lda  length
                        bne  +
                        dec  length+1
+                       dec  length
                        bra  _loop
_done
                    }}
                    repeat separate_pixels {
                        ; TODO optimize this by writing a masked byte in 1 go
                        plot(x, y, color)
                        x++
                    }
                }
                cx16.VERA_ADDR_H &= %00000111   ; vera auto-increment off again
            }
            2,4 -> {
                ; lores 256c
                position(x, y)
                %asm {{
                    lda  color
                    phx
                    ldx  length+1
                    beq  +
                    ldy  #0
-                   sta  cx16.VERA_DATA0
                    iny
                    bne  -
                    dex
                    bne  -
+                   ldy  length     ; remaining
                    beq  +
-                   sta  cx16.VERA_DATA0
                    dey
                    bne  -
+                   plx
                }}
            }
            6 -> {
                ; highres 4c
                ; TODO also mostly usable for lores 4c?
                color &= 3
                ubyte[4] colorbits
                ubyte ii
                for ii in 3 downto 0 {
                    colorbits[ii] = color
                    color <<= 2
                }
                void addr_mul_24_for_highres_4c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                %asm {{
                    lda  cx16.VERA_ADDR_H
                    and  #%00000111         ; no auto advance
                    sta  cx16.VERA_ADDR_H
                    stz  cx16.VERA_CTRL     ; setup vera addr 0
                    lda  cx16.r1
                    and  #1
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0
                    sta  cx16.VERA_ADDR_L
                    lda  cx16.r0+1
                    sta  cx16.VERA_ADDR_M
                    phx
                    ldx  x
                }}

                repeat length {
                    %asm {{
                        txa
                        and  #3
                        tay
                        lda  cx16.VERA_DATA0
                        and  gfx3.plot.mask4c,y
                        ora  colorbits,y
                        sta  cx16.VERA_DATA0
                        cpy  #%00000011         ; next vera byte?
                        bne  ++
                        inc  cx16.VERA_ADDR_L
                        bne  ++
                        inc  cx16.VERA_ADDR_M
+                       bne  +
                        inc  cx16.VERA_ADDR_H
+                       inx                     ; next pixel
                    }}
                }

                %asm {{
                    plx
                }}
            }
        }
    }

    sub vertical_line(uword x, uword y, uword lheight, ubyte color) {
        when active_mode {
            1, 5 -> {
                ; monochrome, lo-res
                cx16.r15L = gfx3.plot.bits[x as ubyte & 7]           ; bitmask
                if color {
                    if monochrome_dont_stipple_flag {
                        ; draw continuous line.
                        position2(x,y,true)
                        if active_mode==1
                            set_both_strides(11)    ; 40 increment = 1 line in 320 px monochrome
                        else
                            set_both_strides(12)    ; 80 increment = 1 line in 640 px monochrome
                        repeat lheight {
                            %asm {{
                                lda  cx16.VERA_DATA0
                                ora  cx16.r15L
                                sta  cx16.VERA_DATA1
                            }}
                        }
                    } else {
                        ; draw stippled line.
                        if x&1 {
                            y++
                            lheight--
                        }
                        position2(x,y,true)
                        if active_mode==1
                            set_both_strides(12)    ; 80 increment = 2 line in 320 px monochrome
                        else
                            set_both_strides(13)    ; 160 increment = 2 line in 640 px monochrome
                        repeat lheight/2 {
                            %asm {{
                                lda  cx16.VERA_DATA0
                                ora  cx16.r15L
                                sta  cx16.VERA_DATA1
                            }}
                        }
                    }
                } else {
                    position2(x,y,true)
                    cx16.r15 = ~cx16.r15    ; erase pixels
                    if active_mode==1
                        set_both_strides(11)    ; 40 increment = 1 line in 320 px monochrome
                    else
                        set_both_strides(12)    ; 80 increment = 1 line in 640 px monochrome
                    repeat lheight {
                        %asm {{
                            lda  cx16.VERA_DATA0
                            and  cx16.r15L
                            sta  cx16.VERA_DATA1
                        }}
                    }
                }
            }
            2,4 -> {
                ; lores 256c
                ; set vera auto-increment to 320 pixel increment (=next line)
                position(x,y)
                cx16.VERA_ADDR_H = cx16.VERA_ADDR_H & %00000111 | (14<<4)
                %asm {{
                    ldy  lheight
                    beq  +
                    lda  color
-                   sta  cx16.VERA_DATA0
                    dey
                    bne  -
+
                }}
            }
            6 -> {
                ; highres 4c
                ; use TWO vera adress pointers simultaneously one for reading, one for writing, so auto-increment is possible
                if lheight==0
                    return
                position2(x,y,true)
                set_both_strides(13)    ; 160 increment = 1 line in 640 px 4c mode
                color &= 3
                color <<= gfx3.plot.shift4c[lsb(x) & 3]         ; TODO with lookup table
                ubyte @shared mask = gfx3.plot.mask4c[lsb(x) & 3]
                repeat lheight {
                    %asm {{
                        lda  cx16.VERA_DATA0
                        and  mask
                        ora  color
                        sta  cx16.VERA_DATA1
                    }}
                }
            }
        }

        sub set_both_strides(ubyte stride) {
            stride <<= 4
            cx16.VERA_CTRL = 0
            cx16.VERA_ADDR_H = cx16.VERA_ADDR_H & %00000111 | stride
            cx16.VERA_CTRL = 1
            cx16.VERA_ADDR_H = cx16.VERA_ADDR_H & %00000111 | stride
        }

    }

    sub line(uword @zp x1, uword @zp y1, uword @zp x2, uword @zp y2, ubyte color) {
        ; Bresenham algorithm.
        ; This code special-cases various quadrant loops to allow simple ++ and -- operations.
        if y1>y2 {
            ; make sure dy is always positive to have only 4 instead of 8 special cases
            cx16.r0 = x1
            x1 = x2
            x2 = cx16.r0
            cx16.r0 = y1
            y1 = y2
            y2 = cx16.r0
        }
        word @zp dx = (x2 as word)-x1
        word @zp dy = (y2 as word)-y1

        if dx==0 {
            vertical_line(x1, y1, abs(dy) as uword +1, color)
            return
        }
        if dy==0 {
            if x1>x2
                x1=x2
            horizontal_line(x1, y1, abs(dx) as uword +1, color)
            return
        }

        word @zp d = 0
        cx16.r13 = true      ; 'positive_ix'
        if dx < 0 {
            dx = -dx
            cx16.r13 = false
        }
        word @zp dx2 = dx*2
        word @zp dy2 = dy*2
        cx16.r14 = x1       ; internal plot X

        if dx >= dy {
            if cx16.r13 {
                repeat {
                    plot(cx16.r14, y1, color)
                    if cx16.r14==x2
                        return
                    cx16.r14++
                    d += dy2
                    if d > dx {
                        y1++
                        d -= dx2
                    }
                }
            } else {
                repeat {
                    plot(cx16.r14, y1, color)
                    if cx16.r14==x2
                        return
                    cx16.r14--
                    d += dy2
                    if d > dx {
                        y1++
                        d -= dx2
                    }
                }
            }
        }
        else {
            if cx16.r13 {
                repeat {
                    plot(cx16.r14, y1, color)
                    if y1 == y2
                        return
                    y1++
                    d += dx2
                    if d > dy {
                        cx16.r14++
                        d -= dy2
                    }
                }
            } else {
                repeat {
                    plot(cx16.r14, y1, color)
                    if y1 == y2
                        return
                    y1++
                    d += dx2
                    if d > dy {
                        cx16.r14--
                        d -= dy2
                    }
                }
            }
        }
    }

    sub circle(uword @zp xcenter, uword @zp ycenter, ubyte radius, ubyte color) {
        ; Midpoint algorithm.
        if radius==0
            return

        ubyte @zp xx = radius
        ubyte @zp yy = 0
        word @zp decisionOver2 = (1 as word)-xx
        ; R14 = internal plot X
        ; R15 = internal plot Y

        while xx>=yy {
            cx16.r14 = xcenter + xx
            cx16.r15 = ycenter + yy
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter - xx
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter + xx
            cx16.r15 = ycenter - yy
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter - xx
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter + yy
            cx16.r15 = ycenter + xx
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter - yy
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter + yy
            cx16.r15 = ycenter - xx
            plot(cx16.r14, cx16.r15, color)
            cx16.r14 = xcenter - yy
            plot(cx16.r14, cx16.r15, color)

            yy++
            if decisionOver2<=0
                decisionOver2 += (yy as word)*2+1
            else {
                xx--
                decisionOver2 += (yy as word -xx)*2+1
            }
        }
    }

    sub disc(uword @zp xcenter, uword @zp ycenter, ubyte @zp radius, ubyte color) {
        ; Midpoint algorithm, filled
        if radius==0
            return
        ubyte @zp yy = 0
        word @zp decisionOver2 = (1 as word)-radius

        while radius>=yy {
            horizontal_line(xcenter-radius, ycenter+yy, radius*$0002+1, color)
            horizontal_line(xcenter-radius, ycenter-yy, radius*$0002+1, color)
            horizontal_line(xcenter-yy, ycenter+radius, yy*$0002+1, color)
            horizontal_line(xcenter-yy, ycenter-radius, yy*$0002+1, color)
            yy++
            if decisionOver2<=0
                decisionOver2 += (yy as word)*2+1
            else {
                radius--
                decisionOver2 += (yy as word -radius)*2+1
            }
        }
    }

    sub plot(uword @zp x, uword y, ubyte color) {
        ubyte[8] @shared bits = [128, 64, 32, 16, 8, 4, 2, 1]
        ubyte[4] @shared mask4c = [%00111111, %11001111, %11110011, %11111100]
        ubyte[4] @shared shift4c = [6,4,2,0]

        when active_mode {
            1 -> {
                ; lores monochrome
                %asm {{
                    lda  x
                    eor  y
                    ora  monochrome_dont_stipple_flag
                    and  #1
                }}
                if_nz {
                    %asm {{
                        lda  x
                        and  #7
                        pha     ; xbits
                    }}
                    x /= 8
                    x += y*(320/8)
                    %asm {{
                        stz  cx16.VERA_CTRL
                        stz  cx16.VERA_ADDR_H
                        lda  x+1
                        sta  cx16.VERA_ADDR_M
                        lda  x
                        sta  cx16.VERA_ADDR_L
                        ply         ; xbits
                        lda  bits,y
                        ldy  color
                        beq  +
                        tsb  cx16.VERA_DATA0
                        bra  ++
+                       trb  cx16.VERA_DATA0
+
                    }}
                }
            }
            ; TODO mode 2,3
            2,4 -> {
                ; lores 256c
                void addr_mul_24_for_lores_256c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                %asm {{
                    stz  cx16.VERA_CTRL
                    lda  cx16.r1
                    ora  #%00010000     ; enable auto-increment so next_pixel() can be used after this
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0+1
                    sta  cx16.VERA_ADDR_M
                    lda  cx16.r0
                    sta  cx16.VERA_ADDR_L
                    lda  color
                    sta  cx16.VERA_DATA0
                }}
            }
            5 -> {
                ; highres monochrome
                %asm {{
                    lda  x
                    eor  y
                    ora  monochrome_dont_stipple_flag
                    and  #1
                }}
                if_nz {
                    %asm {{
                        lda  x
                        and  #7
                        pha     ; xbits
                    }}
                    x /= 8
                    x += y*(640/8)
                    %asm {{
                        stz  cx16.VERA_CTRL
                        stz  cx16.VERA_ADDR_H
                        lda  x+1
                        sta  cx16.VERA_ADDR_M
                        lda  x
                        sta  cx16.VERA_ADDR_L
                        ply     ; xbits
                        lda  bits,y
                        ldy  color
                        beq  +
                        tsb  cx16.VERA_DATA0
                        bra  ++
+                       trb  cx16.VERA_DATA0
+
                    }}
                }
            }
            6 -> {
                ; highres 4c
                ; TODO also mostly usable for lores 4c?
                void addr_mul_24_for_highres_4c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                cx16.r2L = lsb(x) & 3       ; xbits
                color &= 3
                color <<= shift4c[cx16.r2L]     ; TODO with lookup table
                %asm {{
                    stz  cx16.VERA_CTRL
                    lda  cx16.r1L
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0H
                    sta  cx16.VERA_ADDR_M
                    lda  cx16.r0L
                    sta  cx16.VERA_ADDR_L
                    ldy  cx16.r2L           ; xbits
                    lda  mask4c,y
                    and  cx16.VERA_DATA0
                    ora  color
                    sta  cx16.VERA_DATA0
                }}
            }
        }
    }

    sub pget(uword @zp x, uword y) -> ubyte {
        when active_mode {
            1 -> {
                ; lores monochrome
                %asm {{
                    lda  x
                    and  #7
                    pha     ; xbits
                }}
                x /= 8
                x += y*(320/8)
                %asm {{
                    stz  cx16.VERA_CTRL
                    stz  cx16.VERA_ADDR_H
                    lda  x+1
                    sta  cx16.VERA_ADDR_M
                    lda  x
                    sta  cx16.VERA_ADDR_L
                    ply         ; xbits
                    lda  plot.bits,y
                    and  cx16.VERA_DATA0
                    beq  +
                    lda  #1
+
                }}
            }
            ; TODO mode 2 and 3
            2,4 -> {
                ; lores 256c
                void addr_mul_24_for_lores_256c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                %asm {{
                    stz  cx16.VERA_CTRL
                    lda  cx16.r1
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0+1
                    sta  cx16.VERA_ADDR_M
                    lda  cx16.r0
                    sta  cx16.VERA_ADDR_L
                    lda  cx16.VERA_DATA0
                }}
            }
            5 -> {
                ; hires monochrome
                %asm {{
                    lda  x
                    and  #7
                    pha     ; xbits
                }}
                x /= 8
                x += y*(640/8)
                %asm {{
                    stz  cx16.VERA_CTRL
                    stz  cx16.VERA_ADDR_H
                    lda  x+1
                    sta  cx16.VERA_ADDR_M
                    lda  x
                    sta  cx16.VERA_ADDR_L
                    ply     ; xbits
                    lda  plot.bits,y
                    and  cx16.VERA_DATA0
                    beq  +
                    lda  #1
+
                }}
            }
            6 -> {
                ; hires 4c
                void addr_mul_24_for_highres_4c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                %asm {{
                    stz  cx16.VERA_CTRL
                    lda  cx16.r1L
                    sta  cx16.VERA_ADDR_H
                    lda  cx16.r0H
                    sta  cx16.VERA_ADDR_M
                    lda  cx16.r0L
                    sta  cx16.VERA_ADDR_L
                    lda  cx16.VERA_DATA0
                    sta  cx16.r0L
                }}
                cx16.r1L = lsb(x) & 3
                cx16.r0L >>= gfx3.plot.shift4c[cx16.r1L]        ; TODO with lookup table
                return cx16.r0L & 3
            }
            else -> return 0
        }
    }

    sub position(uword @zp x, uword y) {
        ubyte bank
        when active_mode {
            1 -> {
                ; lores monochrome
                cx16.r0 = y*(320/8) + x/8
                cx16.vaddr(0, cx16.r0, 0, 1)
            }
            ; TODO modes 2,3
            2,4 -> {
                ; lores 256c
                void addr_mul_24_for_lores_256c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                bank = lsb(cx16.r1)
                cx16.vaddr(bank, cx16.r0, 0, 1)
            }
            5 -> {
                ; highres monochrome
                cx16.r0 = y*(640/8) + x/8
                cx16.vaddr(0, cx16.r0, 0, 1)
            }
            6 -> {
                ; highres 4c
                void addr_mul_24_for_highres_4c(y, x)      ; 24 bits result is in r0 and r1L (highest byte)
                bank = lsb(cx16.r1)
                cx16.vaddr(bank, cx16.r0, 0, 1)
            }
        }
    }

    sub position2(uword @zp x, uword y, bool also_port_1) {
        position(x, y)
        if also_port_1 {
            when active_mode {
                1, 5 -> cx16.vaddr(0, cx16.r0, 1, 1)
                ; TODO modes 2, 3
                2,4, 6 -> {
                    ubyte bank = lsb(cx16.r1)
                    cx16.vaddr(bank, cx16.r0, 1, 1)
                }
            }
        }
    }

    inline asmsub next_pixel(ubyte color @A) {
        ; -- sets the next pixel byte to the graphics chip.
        ;    for 8 bpp screens this will plot 1 pixel.
        ;    for 1 bpp screens it will plot 8 pixels at once (color = bit pattern).
        ;    for 2 bpp screens it will plot 4 pixels at once (color = bit pattern).
        %asm {{
            sta  cx16.VERA_DATA0
        }}
    }

    asmsub next_pixels(uword pixels @AY, uword amount @R0) clobbers(A, Y)  {
        ; -- sets the next bunch of pixels from a prepared array of bytes.
        ;    for 8 bpp screens this will plot 1 pixel per byte.
        ;    for 1 bpp screens it will plot 8 pixels at once (colors are the bit patterns per byte).
        ;    for 2 bpp screens it will plot 4 pixels at once (colors are the bit patterns per byte).
        %asm {{
            phx
            sta  P8ZP_SCRATCH_W1
            sty  P8ZP_SCRATCH_W1+1
            ldx  cx16.r0+1
            beq  +
            ldy  #0
-           lda  (P8ZP_SCRATCH_W1),y
            sta  cx16.VERA_DATA0
            iny
            bne  -
            inc  P8ZP_SCRATCH_W1+1       ; next page of 256 pixels
            dex
            bne  -

+           ldx  cx16.r0           ; remaining pixels
            beq  +
            ldy  #0
-           lda  (P8ZP_SCRATCH_W1),y
            sta  cx16.VERA_DATA0
            iny
            dex
            bne  -
+           plx
        }}
    }

    asmsub set_8_pixels_from_bits(ubyte bits @R0, ubyte oncolor @A, ubyte offcolor @Y) {
        ; this is only useful in 256 color mode where one pixel equals one byte value.
        %asm {{
            phx
            ldx  #8
-           asl  cx16.r0
            bcc  +
            sta  cx16.VERA_DATA0
            bra  ++
+           sty  cx16.VERA_DATA0
+           dex
            bne  -
            plx
            rts
        }}
    }

    const ubyte charset_bank = $1
    const uword charset_addr = $f000       ; in bank 1, so $1f000

    sub text_charset(ubyte charset) {
        ; -- select the text charset to use with the text() routine
        ;    the charset number is the same as for the cx16.screen_set_charset() ROM function.
        ;    1 = ISO charset, 2 = PETSCII uppercase+graphs, 3= PETSCII uppercase+lowercase.
        cx16.screen_set_charset(charset, 0)
    }

    sub text(uword @zp x, uword y, ubyte color, uword sctextptr) {
        ; -- Write some text at the given pixel position. The text string must be in screencode encoding (not petscii!).
        ;    You must also have called text_charset() first to select and prepare the character set to use.
        ;    NOTE: in monochrome (1bpp) screen modes, x position is currently constrained to multiples of 8 !  TODO allow per-pixel horizontal positioning
        ; TODO draw whole horizontal spans using vera auto increment if possible, instead of per-character columns
        uword chardataptr
        when active_mode {
            1, 5 -> {
                ; monochrome mode, either resolution
                cx16.r2 = 40
                if active_mode==5
                    cx16.r2 = 80
                while @(sctextptr) {
                    chardataptr = charset_addr + (@(sctextptr) as uword)*8
                    cx16.vaddr(charset_bank, chardataptr, 1, 1)
                    position(x,y)
                    %asm {{
                        lda  cx16.VERA_ADDR_H
                        and  #%111              ; don't auto-increment, we have to do that manually because of the ora
                        sta  cx16.VERA_ADDR_H
                        lda  color
                        sta  P8ZP_SCRATCH_B1
                        ldy  #8
-                       lda  P8ZP_SCRATCH_B1
                        bne  +                  ; white color, plot normally
                        lda  cx16.VERA_DATA1
                        eor  #255               ; black color, keep only the other pixels
                        and  cx16.VERA_DATA0
                        bra  ++
+                       lda  cx16.VERA_DATA0
                        ora  cx16.VERA_DATA1
+                       sta  cx16.VERA_DATA0
                        lda  cx16.VERA_ADDR_L
                        clc
                        adc  cx16.r2
                        sta  cx16.VERA_ADDR_L
                        bcc  +
                        inc  cx16.VERA_ADDR_M
+                       inc  x
                        bne  +
                        inc  x+1
+                       dey
                        bne  -
                    }}
                    sctextptr++
                }
            }
            4 -> {
                ; lores 256c
                while @(sctextptr) {
                    chardataptr = charset_addr + (@(sctextptr) as uword)*8
                    cx16.vaddr(charset_bank, chardataptr, 1, 1)
                    repeat 8 {
                        position(x,y)
                        y++
                        %asm {{
                            phx
                            ldx  color
                            lda  cx16.VERA_DATA1
                            sta  P8ZP_SCRATCH_B1
                            ldy  #8
-                           asl  P8ZP_SCRATCH_B1
                            bcc  +
                            stx  cx16.VERA_DATA0    ; write a pixel
                            bra  ++
+                           lda  cx16.VERA_DATA0    ; don't write a pixel, but do advance to the next address
+                           dey
                            bne  -
                            plx
                        }}
                    }
                    x+=8
                    y-=8
                    sctextptr++
                }
            }
            6 -> {
                ; hires 4c
                ; we're going to use a few cx16 registers to make sure every variable is in zeropage in the inner loop.
                cx16.r11L = color
                cx16.r8 = y
                while @(sctextptr) {
                    chardataptr = charset_addr + (@(sctextptr) as uword)*8
                    repeat 8 {
                        ; TODO rewrite this inner loop partly in assembly
                        ;      requires expanding the charbits to 2-bits per pixel (based on color)
                        ;      also it's way more efficient to draw whole horizontal spans instead of per-character
                        cx16.r9L = cx16.vpeek(charset_bank, chardataptr)  ; get the 8 horizontal character bits
                        cx16.r7 = x
                        repeat 8 {
                            cx16.r9L <<= 1
                            if_cs
                                plot(cx16.r7, cx16.r8, cx16.r11L)
                            cx16.r7++
                        }
                        chardataptr++
                        cx16.r8++
                    }
                    x+=8
                    cx16.r8-=8
                    sctextptr++
                }
            }
        }
    }

    asmsub cs_innerloop640() clobbers(Y) {
        %asm {{
            ldy  #80
-           stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            stz  cx16.VERA_DATA0
            dey
            bne  -
            rts
        }}
    }

    asmsub addr_mul_24_for_highres_4c(uword yy @R2, uword xx @R3)  clobbers(A, Y) -> uword @R0, uword @R1 {
        ; yy * 160 + xx/4  (24 bits calculation)
        ; 24 bits result is in r0 and r1L (highest byte)
        %asm {{
            ldy  #5
-           asl  cx16.r2
            rol  cx16.r2+1
            dey
            bne  -
            lda  cx16.r2
            sta  cx16.r0
            lda  cx16.r2+1
            sta  cx16.r0+1
            asl  cx16.r0
            rol  cx16.r0+1
            asl  cx16.r0
            rol  cx16.r0+1

            ; xx >>= 2  (xx=R3)
            lsr  cx16.r3+1
            ror  cx16.r3
            lsr  cx16.r3+1
            ror  cx16.r3

            ; add r2 and xx (r3) to r0 (24-bits)
            stz  cx16.r1
            clc
            lda  cx16.r0
            adc  cx16.r2
            sta  cx16.r0
            lda  cx16.r0+1
            adc  cx16.r2+1
            sta  cx16.r0+1
            bcc  +
            inc  cx16.r1
+           clc
            lda  cx16.r0
            adc  cx16.r3
            sta  cx16.r0
            lda  cx16.r0+1
            adc  cx16.r3+1
            sta  cx16.r0+1
            bcc  +
            inc  cx16.r1
+
            rts
        }}
    }

    asmsub addr_mul_24_for_lores_256c(uword yy @R0, uword xx @AY) clobbers(A) -> uword @R0, ubyte @R1  {
        ; yy * 320 + xx (24 bits calculation)
        %asm {{
            sta  P8ZP_SCRATCH_W1
            sty  P8ZP_SCRATCH_W1+1
            lda  cx16.r0
            sta  P8ZP_SCRATCH_B1
            lda  cx16.r0+1
            sta  cx16.r1
            sta  P8ZP_SCRATCH_REG
            lda  cx16.r0
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            asl  a
            rol  P8ZP_SCRATCH_REG
            sta  cx16.r0
            lda  P8ZP_SCRATCH_B1
            clc
            adc  P8ZP_SCRATCH_REG
            sta  cx16.r0+1
            bcc  +
            inc  cx16.r1
+           ; now add the value to this 24-bits number
            lda  cx16.r0
            clc
            adc  P8ZP_SCRATCH_W1
            sta  cx16.r0
            lda  cx16.r0+1
            adc  P8ZP_SCRATCH_W1+1
            sta  cx16.r0+1
            bcc  +
            inc  cx16.r1
+           lda  cx16.r1
            rts
        }}
    }

}
